# Tutorial: User-defined matrices - boundary element method

Suppose you have a new type of NEP, which does not naturally fit
into the standard types in NEP-PACK. This tutorial shows how
you can define a NEP where the only way to access the NEP
is a function to compute ``M^{(k)}(λ)``. To illustrate this
we use a boundary element method approach for computation
of resonances. The complete code is available in
`gallery_extra/bem_hardcoded`. The example is also
available as a gallery problem: `nep=nep_gallery("bem_fichera")`.

## Boundary element method

The boundary element method applied to Helmholtz eigenvalue
problem can be described by the matrix consisting of elements
```math
[M(λ)]_{ij}=\frac{1}{4\pi}\int_{\Delta_i}\int_{\Delta_j}\frac{e^{\iota\lambda\|\xi-\eta\|}}{\|\xi-\eta\|}dS(\eta)dS(\xi)
```
where $\Delta_i$,$i=1,\ldots,n$ are boundary elements.
The boundary element approach is available through three functions
```julia
mesh=gen_ficheramesh(N) # computes a mesh
precompute_quad!(mesh,gauss_order) # precompute quadrature coefficients
assemble_BEM(λ, mesh, gauss_order,der) # Compute the matrix consisting of all the integrals corresponding to λ
```
These functions are based on the model (and inspired by some of the code) in
in ["A boundary element method for solving PDE eigenvalue problems", Steinlechner, bachelor thesis, ETH Zürich, 2010](http://sma.epfl.ch/~anchpcommon/students/steinlechner.pdf) and also
used in the simulations in
["Chebyshev interpolation for nonlinear eigenvalue problems", Effenberger, Kressner, BIT Numerical Mathematics, 2012, Volume 52, Issue 4, pp 933–951](https://link.springer.com/article/10.1007/s10543-012-0381-5).

## Implementation in NEP-PACK

In order to define your new NEP you need to define a new NEP-type
```julia
struct BEM_NEP <: NEP
    mesh::Vector{Triangle}
    n::Int
    gauss_order::Int
end
```
The `mesh` variable is a vector of triangle objects defining the domain,
`n` is the size of the mesh and `gauss_order` the quadrature order.
All NEPs have to defined `size()` functions
```julia
# We overload the size function from Base so we need to import it explicitly
import Base.size;
function size(nep::BEM_NEP)
    return (nep.n,nep.n);
end
function size(nep::BEM_NEP,dim)
    return nep.n;
end
```
The function `assemble_BEM` computes the matrix defined by the integrals. Hence,
we need to call this function for every call to `compute_Mder`:
```julia
import NonlinearEigenproblems.NEPCore.compute_Mder # We overload the function
function compute_Mder(nep::BEM_NEP,λ::Number,der::Int=0)
    return assemble_BEM(λ, nep.mesh, nep.gauss_order,der)[:,:,1];
end
```
In order to make other compute functions available to the methods,
we can use the conversion functions. In particular, the `compute_Mlincomb` function
can be implemented by making several calls in `compute_Mder`. This
is done in the NEP-PACK-provided helper function `compute_Mlincomb_from_Mder`.
We make this the default behaviour for this NEP:
```julia
import NonlinearEigenproblems.NEPCore.compute_Mlincomb # Since we overload
# Delegate the compute Mlincomb functions. This can be quite inefficient.
compute_Mlincomb(nep::BEM_NEP,λ::Number,V::AbstractVecOrMat, a::Vector) =
      compute_Mlincomb_from_Mder(nep,λ,V,a)
compute_Mlincomb(nep::BEM_NEP,λ::Number,V::AbstractVecOrMat) =
      compute_Mlincomb(nep,λ,V, ones(eltype(V),size(V,2)))
```
We can now create a `BEM_NEP` as follows:
```julia
gauss_order=3; N=5;
mesh=gen_ficheramesh(5)
precompute_quad!(mesh,gauss_order)
nep=BEM_NEP(mesh,gauss_order);
```
## Solving the NEP
After creating the NEP, you can try to solve the problem with methods in the package, e.g.,
[`MSLP`](methods.md#NonlinearEigenproblems.NEPSolver.mslp) works quite well for this problem:
```julia-repl
julia> (λ,v)=mslp(nep,λ=8,displaylevel=1)
Iteration:1 errmeasure:4.122635537095636e-6 λ=8.128272919317748 + 0.007584851218214716im
Iteration:2 errmeasure:1.787963303973586e-8 λ=8.132181234599427 - 1.952792817964521e-5im
Iteration:3 errmeasure:3.2884958163572594e-13 λ=8.132145310156643 - 1.2648247028455485e-5im
Iteration:4 errmeasure:4.6607986030841e-18 λ=8.132145310195453 - 1.264891804832194e-5im
(8.132145310195453 - 1.264891804832194e-5im, Complex{Float64}[3.08473e-5-9.8713e-6im, 9.46458e-5+2.08586e-5im, -0.000418303-9.3624e-5im, -2.27161e-5+3.2045e-5im, -0.00168228-0.000446522im, -0.00660488-0.0018462im, -0.00705554-0.00195021im, -0.000714245-0.000123651im, -0.010653-0.00296256im, -0.0250155-0.00702815im  …  0.00369925+0.00101557im, 0.025547+0.00717101im, 0.0333126+0.00931856im, 0.0158614+0.00438927im, 0.00325204+0.000835354im, 0.021329+0.00595943im, 0.0126512+0.0034611im, 0.00130882+0.000172086im, 0.00131286+0.000207463im, 0.0125435+0.00344975im])
```
This is the computed solution
```@raw html
<br>
<img src="https://user-images.githubusercontent.com/11163595/49595409-324b7d80-f978-11e8-818d-eeeaf9441505.png" height=450>
```


The plotting was done with the following code (by using internals of the BEM-implementation):
```julia
using NonlinearEigenproblems, PyPlot
nep=nep_gallery("bem_fichera")
(λ,v)=mslp(nep,λ=8.1,displaylevel=1)
v=v./maximum(abs.(v));
for k=1:size(nep.mesh,1);
    tri=nep.mesh[k];
    col=[1-abs.(v)[k];0;0]; # plot abslolute value
    X=[tri.P1[1] tri.P2[1]; tri.P3[1] tri.P3[1]];
    Y=[tri.P1[2] tri.P2[2]; tri.P3[2] tri.P3[2]];
    Z=[tri.P1[3] tri.P2[3]; tri.P3[3] tri.P3[3]];
    plot_surface(X,Y,Z,color=col,alpha=0.8);
    plot_wireframe(X,Y,Z,color=[0;0;0],linewidth=1,alpha=0.5,);
end
```
